import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras.metrics import Metric


def update_metric(x_prev, x_now, i):
    return (1 / (i + 1)) * (i * x_prev + x_now)


class IndexAccHist(Metric):
    def __init__(self, name='index_accuracy_histogram'):
        super(IndexAccHist, self).__init__()
        self.metric_name = name
        self.metric = []
        self.tics = self.add_weight(name='tics', initializer="zeros")

    def update_state(self, y_true, y_pred, x=None, sample_weight=None):
        y_pred = tf.argmax(y_pred, axis=-1, output_type=tf.int32)
        equals = tf.cast(y_pred == y_true, tf.float32)
        new_index_accs = K.sum(equals, axis=0) / equals.shape[0]
        for i in range(new_index_accs.shape[0]):
            if i >= len(self.metric):
                self.metric.append(self.add_weight(name=self.metric_name + str(i), initializer="zeros"))
            self.metric[i].assign(update_metric(self.metric[i], new_index_accs[i], self.tics))

        self.tics.assign_add(1.)

    def result(self):
        return self.metric

    def reset(self):
        for i in range(len(self.metric)):
            self.metric[i].assign(0.)

        self.tics.assign(0.)



from tensorflow.keras import layers
from tensorflow import keras


class TransformerBlock(layers.Layer):
    def __init__(self, embed_dim, num_heads, ff_dim, dropout=0.1, ff_activation='relu'):
        super(TransformerBlock, self).__init__()
        # TODO: resolve (embed_dim // num_heads) vs (embed_dim)
        self.att = layers.MultiHeadAttention(num_heads=num_heads, key_dim=embed_dim)
        self.ffn = keras.Sequential(
            [layers.Dense(ff_dim, activation=ff_activation), layers.Dense(embed_dim), ]
        )
        self.layernorm1 = layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = layers.LayerNormalization(epsilon=1e-6)
        self.dropout1 = layers.Dropout(dropout)
        self.dropout2 = layers.Dropout(dropout)

    def call(self, inputs, training=False):
        attn_output = self.att(inputs, inputs)
        attn_output = self.dropout1(attn_output, training=training)
        out1 = self.layernorm1(inputs + attn_output)
        ffn_output = self.ffn(out1)
        ffn_output = self.dropout2(ffn_output, training=training)
        return self.layernorm2(out1 + ffn_output)



import numpy as np
import tensorflow as tf

from .sentence_piece import data_gen, random_iter
from .sentence_piece_termostat import data_gen_termostat


def data_gen_svec(data_path, tokenizer, batch_size, max_len, mask_prob=0.15, overfit_batch=False, helper_mode='none',
                  once=False, session_steps=-1, skip_shorter_than=None):
    gen = data_gen(data_path, tokenizer, batch_size, max_len, mask_prob, overfit_batch, helper_mode, once,
                   session_steps, skip_shorter_than)

    while True:
        tokens_in, tokens_out, mask = next(gen)

        yield ({"tokens": tokens_in, "mask": mask},
               {"deemb": tokens_out, "emb": tokens_out},
               )


def data_gen_termostat_svec(data_path, tokenizer, batch_size, max_len, hint_callback, mask_prob=0.15):
    gen = data_gen_termostat(data_path, tokenizer, batch_size, max_len, hint_callback, mask_prob)

    while True:
        tokens_in, tokens_out, mask = next(gen)

        yield ({"tokens": tokens_in, "mask": mask},
               {"deemb": tokens_out, "emb": tokens_out},
               mask
               )


def random_iter_svec(batch_size, max_len, mask_prob, helper_mode='none', session_steps=-1):
    gen = random_iter(batch_size, max_len, mask_prob, helper_mode, session_steps)

    while True:
        tokens_in, tokens_out = next(gen)

        bert_mask = tf.logical_or(tokens_in == TOKENIZER.mask_token_id, tokens_in > len(TOKENIZER.vocab))

        yield ({"tokens": tokens_in, "mask": bert_mask},
               {"deemb": tokens_out, "emb": tokens_out},
               )



from tensorflow.keras.layers import Conv1D, BatchNormalization, Dropout, Input, Activation, Conv1DTranspose
from tensorflow.keras.models import Model

from prot2vec.layers.skip_block_conv1d import conv1D_conv_block, conv1D_deconv_block, conv1D_identity_block


def conv1d_model(seq_len, onehot_size, dropout=0.1):
    x_input = Input((seq_len, onehot_size))

    x = Conv1D(128, kernel_size=9, strides=2, padding='same')(x_input)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    x = Dropout(dropout)(x)

    # Define size of sub-blocks and initial filter size
    block_layers = [2, 3, 3, 2]
    kernels = 128
    # Step 3 Add the Resnet Blocks
    for i in range(4):
        if i == 0:
            # For sub-block 1 Residual/Convolutional block not needed
            for j in range(block_layers[i]):
                x = conv1D_identity_block(x, kernels, 5)
        else:
            # One Residual/Convolutional Block followed by Identity blocks
            # The filter size will go on increasing by a factor of 2
            kernels = kernels * 2
            x = conv1D_conv_block(x, kernels, 5)
            for j in range(block_layers[i] - 1):
                x = conv1D_identity_block(x, kernels, 5)

    #############
    # Define size of sub-blocks and initial filter size
    block_layers = [2, 3, 3]
    kernels = 1024
    # Step 3 Add the Resnet Blocks
    for i in range(3):
        kernels = kernels / 2
        x = conv1D_deconv_block(x, kernels, 5)
        for j in range(block_layers[i] - 1):
            x = conv1D_identity_block(x, kernels, 5)

    x = Conv1DTranspose(onehot_size, 9, padding='same', strides=2)(x)
    ae_out = Activation('softmax', name='ae_out')(x)

    model = Model(inputs=x_input, outputs=ae_out, name="conv1d_model")

    return model



import json
from pathlib import Path

import numpy as np
from tensorflow.keras.callbacks import Callback


class HintTermostat(Callback):
    def __init__(self, tokenizer_path, target_value, log_path, init_step=0, verbose=False, target_variable='loss',
                 is_min_problem=True):
        super(HintTermostat, self).__init__()
        self.levels = np.abs(np.load(Path(tokenizer_path) / 'termostat.npy'))
        with open(Path(tokenizer_path) / 'hints.json', 'r', encoding='utf-8') as fr:
            self.hints = json.load(fr)
        self.level_num = init_step
        self.target_value = target_value
        self.log_path = Path(log_path)
        self.verbose = verbose
        self.target_variable = target_variable

        # True if the objective is minimizing the target value, False if objective is maximizing the target value
        self.is_min_problem = is_min_problem
        if is_min_problem:
            self.target_value *= -1

    def get_distribution(self):
        return self.levels[self.level_num]

    def on_epoch_end(self, epoch, logs=None):
        actual_value = logs[self.target_variable]
        if not self.is_min_problem:
            actual_value *= -1
        if actual_value < self.target_value:
            self.level_num += 1
        else:
            self.level_num -= 1

        # Saturate by range of values
        self.level_num = max(self.level_num, 0)
        self.level_num = min(self.level_num, self.levels.shape[0])
        if self.verbose:
            print(f'Level: {self.level_num}/{self.levels.shape[0]}')
        with open(self.log_path / 'levels.txt', 'a', encoding='utf-8') as fw:
            print(f'{self.level_num}', file=fw)



import numpy as np
from copy import deepcopy
import json
import datetime as dt
import pickle

from ..analysis.index_accuracy import (
    check_lengths,
    extract_results as extract_results_indexes,
)
from ..analysis.sentencepiece_and_protein_accuracy import (
    extract_results as extract_results_sp_prot,
)


def vectorize_seq(seq, max_len, tokenizer):
    inputs = tokenizer.encode_plus(
        seq,
        add_special_tokens=True,
        max_length=max_len,
        padding="max_length",
        truncation=True,
    )
    input_ids = inputs["input_ids"]
    return input_ids


def vectorize_batch_same_mask_position(batch, mask_position, max_len, tokenizer):
    tokens_in = []

    for (t_in,) in batch:
        t_in = vectorize_seq(t_in, max_len, tokenizer)

        tokens_in.append(t_in)

    tokens_in = np.array(tokens_in, dtype=np.int32)
    mask = np.zeros(tokens_in.shape)

    tokens_masked = deepcopy(tokens_in)

    masked_indexes = [[mask_position] for _ in range(tokens_in.shape[0])]
    masked_coord_y = np.concatenate(masked_indexes)
    masked_coord_x = np.repeat(np.arange(tokens_in.shape[0]), 1)

    tokens_masked[masked_coord_x, masked_coord_y] = tokenizer.mask_token_id

    mask[masked_coord_x, masked_coord_y] = 1

    return tokens_masked, tokens_in, mask


def vectorize_batch_diff_mask_position(batch, masked_positions, max_len, tokenizer):
    tokens_in = []
    domain_linker = list()

    for ((t_in,), dom_lin) in batch:
        t_in = vectorize_seq(t_in, max_len, tokenizer)

        tokens_in.append(t_in)
        domain_linker.append(dom_lin)

    tokens_in = np.array(tokens_in, dtype=np.int32)
    mask = np.zeros(tokens_in.shape)

    tokens_masked = deepcopy(tokens_in)

    masked_indexes = [[masked_positions[i]] for i in range(tokens_in.shape[0])]
    masked_coord_y = np.concatenate(masked_indexes)
    masked_coord_x = np.repeat(np.arange(tokens_in.shape[0]), 1)

    tokens_masked[masked_coord_x, masked_coord_y] = tokenizer.mask_token_id

    mask[masked_coord_x, masked_coord_y] = 1

    return tokens_masked, tokens_in, mask, domain_linker


def get_protein_dom_link_info(key, tokens, seq, domains_info):

    domains = domains_info.get(key)
    if not domains:
        return ["U"] * len(tokens)

    link_dom_sp_vec = ["L"] * len(tokens)
    link_dom_aa_vec = ["L"] * len(seq)

    for domain in domains:
        for i in range(int(domain[1]), int(domain[2]) + 1):
            if i < len(link_dom_aa_vec):
                link_dom_aa_vec[i] = "D"

    actual_index = 0
    for i, token in enumerate(tokens):
        if link_dom_aa_vec[actual_index] == "D":
            link_dom_sp_vec[i] = "D"
        actual_index += len(token)

    return link_dom_sp_vec


def process_all_data(
    model,
    dataset_path,
    tokenizer,
    max_len,
    dev_size,
    results_path_sp,
    results_path_proteins,
    results_path_confusion,
    domains_info,
    save_every_nth=1000,
    print_every_nth=100,
    to_process=None,
    min_mask_index=4,
    matrix_identity=None,
    matrix_similarity=None,
    matrix_set_similarity=None,
    matrix_cosine_similarity=None,
):
    results_sp = dict()
    conf_matrix = dict()
    start_token = 5
    for i, key in enumerate(
        {k: v for k, v in sorted(tokenizer.vocab.items(), key=lambda item: item[1])}
    ):
        if i >= start_token:
            results_sp[key] = {
                "D": [0, 0, 0, 0, 0, 0, 0, 0],
                "L": [0, 0, 0, 0, 0, 0, 0, 0],
                "U": [0, 0, 0, 0, 0, 0, 0, 0],
            }
            conf_matrix[key] = [0] * (len(tokenizer.vocab) - start_token)
    inverted_vocab = {v: k for k, v in tokenizer.vocab.items()}

    max_len1 = max_len - 2
    processed = 0
    middle = max_len // 2
    max_mask_index = max_len - min_mask_index

    with open(dataset_path, "r", encoding="utf-8") as fr:

        batch = []
        masked_positions = []
        for line in fr:
            items = line.strip().split("\t")
            id, seq_id, seqF = items
            if id.endswith(";"):
                id = id[:-1]

            key = f"{id}-{seq_id}"

            tokens = tokenizer.tokenize(seqF)

            if len(tokens) < max_len1:
                continue

            dom_link_info = get_protein_dom_link_info(key, tokens, seqF, domains_info)
            results_protein = [list(), list()]

            offset = 0
            mask_index = min_mask_index
            while True:

                tokens1 = tokens[offset : offset + max_len1]
                seq1 = "".join(tokens1)

                batch.append(((seq1,), dom_link_info[offset + mask_index - 1]))
                masked_positions.append(mask_index)
                if len(batch) >= dev_size:
                    results_sp, results_protein, conf_matrix = extract_results_sp_prot(
                        model,
                        vectorize_batch_diff_mask_position(
                            batch, masked_positions, max_len, tokenizer
                        ),
                        masked_positions,
                        results_sp,
                        results_protein,
                        conf_matrix,
                        inverted_vocab,
                        matrix_identity,
                        matrix_similarity,
                        matrix_set_similarity,
                        matrix_cosine_similarity,
                        start_token,
                    )
                    del batch[:]
                    del masked_positions[:]

                if mask_index < middle:
                    mask_index += 1
                elif offset + max_len1 == len(tokens):
                    if mask_index < max_mask_index:
                        mask_index += 1
                    else:
                        break
                else:
                    offset += 1
            if batch:
                results_sp, results_protein, conf_matrix = extract_results_sp_prot(
                    model,
                    vectorize_batch_diff_mask_position(
                        batch, masked_positions, max_len, tokenizer
                    ),
                    masked_positions,
                    results_sp,
                    results_protein,
                    conf_matrix,
                    inverted_vocab,
                    matrix_identity,
                    matrix_similarity,
                    matrix_set_similarity,
                    matrix_cosine_similarity,
                    start_token,
                )
                del batch[:]
                del masked_positions[:]

            processed += 1

            with open(results_path_proteins, "ab") as af:
                prot_data = f"{key};{','.join(map(str, results_protein[0]))};{','.join(map(str, results_protein[1]))}\n"
                af.write(str.encode(prot_data))

            if processed % print_every_nth == 0:
                print(processed, end=", ")

            if processed % save_every_nth == 0:

                print(f"\nprocessed: {processed}, {dt.datetime.now()}", end=", ")
                with open(results_path_sp, "w") as wf:
                    json.dump(results_sp, wf)
                print(dt.datetime.now(), end=", ")
                with open(results_path_confusion, "wb") as wf:
                    pickle.dump(conf_matrix, wf)
                print(dt.datetime.now())

            if to_process and processed == to_process:
                break


def process_data_indexes(
    model,
    dataset_path,
    tokenizer,
    max_len,
    dev_size,
    results_path,
    save_every_nth=10,
    to_process=1000,
):
    mask_positions = range(1, max_len - 1)

    results = dict()
    for index in mask_positions:
        results[index] = (list(), list())

    max_len1 = max_len - 2
    processed = 0

    with open(dataset_path, "r", encoding="utf-8") as fr:

        batch = []
        for line in fr:
            items = line.strip().split("\t")
            _, _, seqF = items

            tokens = tokenizer.tokenize(seqF)

            if len(tokens) < max_len1 * 2 + 1:
                continue

            for mask_position in mask_positions:
                offset = max_len1 - mask_position
                border = len(tokens) - max_len1 + offset
                while True:

                    if offset + max_len1 >= border:
                        break

                    tokens1 = tokens[offset : offset + max_len1]
                    assert len(tokens1) == max_len1
                    seq1 = "".join(tokens1)

                    batch.append((seq1,))
                    if len(batch) >= dev_size:
                        results = extract_results_indexes(
                            model,
                            vectorize_batch_same_mask_position(
                                batch, mask_position, max_len, tokenizer
                            ),
                            mask_position,
                            results,
                        )
                        del batch[:]
                    offset += 1
                if batch:
                    results = extract_results_indexes(
                        model,
                        vectorize_batch_same_mask_position(
                            batch, mask_position, max_len, tokenizer
                        ),
                        mask_position,
                        results,
                    )
                    del batch[:]

            processed += 1

            print(processed, end=", ")

            if processed % save_every_nth == 0:
                check_lengths(results)
                print(f"\nprocessed: {processed}")
                with open(results_path, "w") as wf:
                    json.dump(results, wf)

            if to_process and processed == to_process:
                break


def read_architectures(datapath):
    # protein-id: [(domain name, start, end), ...]

    # read architectures
    ret = {}
    with open(datapath, "r", encoding="utf-8") as rf:
        for line in rf:
            items = line.strip().split("\t")
            # print(items)
            key = f"{items[0]}-{items[1]}"
            arch = [tuple(items[i : i + 3]) for i in range(2, len(items), 3)]
            ret[key] = arch
    return ret



import json
from matplotlib import pyplot as plt
import numpy as np

from tensorflow.keras import backend as K
from transformers.modeling_tf_outputs import TFMaskedLMOutput


def extract_results(model, data, mask_position, results: dict):

    predictions = model.predict(data[0])
    if isinstance(predictions, TFMaskedLMOutput):
        predictions = K.softmax(predictions.logits, axis=-1)
    predicted_tokens = K.argmax(predictions, axis=-1)

    for i, prediction in enumerate(predictions):
        predicted_token = int(predicted_tokens[i][mask_position])
        true_token = data[1][i][mask_position]
        soft_predicted = float(
            prediction[mask_position][predicted_token]
        )  # softmax_out
        soft_true = float(prediction[mask_position][true_token])  # softmax_out
        results[mask_position][0].append(soft_predicted)
        results[mask_position][1].append(soft_true)
    return results


def check_lengths(data):
    lengths = list()
    for vec1, _ in data.values():
        lengths.append(len(vec1))
    len_0 = lengths[0]
    for length in lengths:
        if length != len_0:
            raise ValueError("Lengths mismatch")


def show_results(
    results_path,
    show_confidences: bool = True,
    show_norm_confidences: bool = True,
    show_accuracies: bool = True,
    show_diff_from_mean: bool = True,
    show_diff_on_indexes: bool = True,
):
    with open(results_path, "r") as rf:
        results = json.load(rf)

    pred_avgs = list()
    true_avgs = list()

    for i in range(len(list(results.values())[0][0])):
        i_val_pred = [val[0][i] for val in results.values()]
        i_val_true = [val[1][i] for val in results.values()]

        pred_avgs.append(sum(i_val_pred) / len(i_val_pred))
        true_avgs.append(sum(i_val_true) / len(i_val_true))

    ratios_mean = list()
    trues_mean = list()
    preds_mean = list()

    ratios_var = list()
    trues_var = list()
    preds_var = list()

    hits = list()

    trues_diff_avg = list()
    preds_diff_avg = list()

    for _, values in results.items():
        vec_preds, vec_trues = values
        vec_ratios = [v_t / v_p for v_p, v_t in zip(vec_preds, vec_trues)]
        mean_ratio = sum(vec_ratios) / len(vec_ratios)
        mean_true = sum(vec_trues) / len(vec_trues)
        mean_pred = sum(vec_preds) / len(vec_preds)

        ratios_mean.append(mean_ratio)
        trues_mean.append(mean_true)
        preds_mean.append(mean_pred)

        ratios_var.append(
            sum([(v_r - mean_ratio) ** 2 for v_r in vec_ratios]) / len(vec_ratios)
        )
        trues_var.append(
            sum([(v_t - mean_true) ** 2 for v_t in vec_trues]) / len(vec_trues)
        )
        preds_var.append(
            sum([(v_p - mean_pred) ** 2 for v_p in vec_preds]) / len(vec_preds)
        )

        vec_hits = [int(v) for v in vec_ratios]
        hits.append(sum(vec_hits) / len(vec_hits))

        trues_diff = [abs(m - v) for m, v in zip(true_avgs, vec_trues)]
        preds_diff = [abs(m - v) for m, v in zip(pred_avgs, vec_preds)]

        trues_diff_avg.append(sum(trues_diff) / len(trues_diff))
        preds_diff_avg.append(sum(preds_diff) / len(preds_diff))

    if show_confidences:
        plt.rcParams["figure.figsize"] = (30, 5)

        plt.plot(
            np.arange(0, len(ratios_mean)), ratios_mean, color="black", linewidth=2
        )
        plt.plot(np.arange(0, len(preds_mean)), preds_mean, color="blue", linewidth=2)
        plt.plot(np.arange(0, len(trues_mean)), trues_mean, color="red", linewidth=2)

        plt.fill_between(
            np.arange(0, len(ratios_mean)),
            [m - v for m, v in zip(ratios_mean, ratios_var)],
            [m + v for m, v in zip(ratios_mean, ratios_var)],
            color="yellow",
        )
        plt.fill_between(
            np.arange(0, len(preds_mean)),
            [m - v for m, v in zip(preds_mean, preds_var)],
            [m + v for m, v in zip(preds_mean, preds_var)],
            color="orange",
        )
        plt.fill_between(
            np.arange(0, len(trues_mean)),
            [m - v for m, v in zip(trues_mean, trues_var)],
            [m + v for m, v in zip(trues_mean, trues_var)],
            color="green",
        )

        plt.title("averages on indexes")
        plt.xlabel("position in window")
        plt.legend(["ratio", "pred", "true", "var", "var", "var"])
        plt.xticks(np.arange(0, len(ratios_mean), 1))
        plt.grid()
        plt.show()

    norm_true = list()
    norm_pred = list()
    max_true = max(trues_mean)
    max_pred = max(preds_mean)

    if show_norm_confidences:
        for v_t, v_p in zip(trues_mean, preds_mean):
            norm_true.append(v_t / max_true)
            norm_pred.append(v_p / max_pred)

        plt.rcParams["figure.figsize"] = (30, 5)

        plt.plot(np.arange(0, len(norm_pred)), norm_pred, color="blue", linewidth=2)
        plt.plot(np.arange(0, len(norm_true)), norm_true, color="red", linewidth=2)

        plt.title("norm averages on indexes")
        plt.xlabel("position in window")
        plt.legend(["pred", "true"])
        plt.xticks(np.arange(0, len(norm_pred), 1))
        plt.grid()
        plt.show()

    if show_accuracies:
        plt.rcParams["figure.figsize"] = (30, 5)
        plt.bar(np.arange(0, len(hits)), hits, color="black", linewidth=2)
        plt.title("acurracy on indexes")
        plt.xlabel("position in window")
        plt.legend(["accuracy"])
        plt.xticks(np.arange(0, len(hits), 1))
        plt.grid()
        plt.show()

    if show_diff_from_mean:
        plt.rcParams["figure.figsize"] = (30, 5)

        plt.plot(
            np.arange(0, len(preds_diff_avg)), preds_diff_avg, color="blue", linewidth=2
        )
        plt.plot(
            np.arange(0, len(trues_diff_avg)), trues_diff_avg, color="red", linewidth=2
        )

        plt.title(f"differences from mean values")
        plt.xlabel("position in window")
        plt.legend(["pred", "true"])
        plt.xticks(np.arange(0, len(trues_diff_avg), 1))
        plt.grid()
        plt.show()

    if show_diff_on_indexes:
        for key_ref, values_ref in results.items():

            trues_diffs = list()
            pred_diffs = list()
            ratios_diffs = list()

            ref_true = values_ref[1]
            ref_pred = values_ref[0]
            ref_ratio = [t / p for p, t in zip(ref_pred, ref_true)]

            for key, values in results.items():
                vec_preds, vec_trues = values
                vec_ratios = [v_t / v_p for v_p, v_t in zip(vec_preds, vec_trues)]

                # differences
                diff_true_vec = [abs(m - v) for m, v in zip(ref_true, vec_trues)]
                diff_pred_vec = [abs(m - v) for m, v in zip(ref_pred, vec_preds)]
                diff_ratio_vec = [abs(m - v) for m, v in zip(ref_ratio, vec_ratios)]

                trues_diffs.append(sum(diff_true_vec) / len(diff_true_vec))
                pred_diffs.append(sum(diff_pred_vec) / len(diff_pred_vec))
                ratios_diffs.append(sum(diff_ratio_vec) / len(diff_ratio_vec))

            plt.rcParams["figure.figsize"] = (30, 5)

            plt.plot(
                np.arange(0, len(ratios_diffs)),
                ratios_diffs,
                color="black",
                linewidth=2,
            )
            plt.plot(
                np.arange(0, len(pred_diffs)), pred_diffs, color="blue", linewidth=2
            )
            plt.plot(
                np.arange(0, len(trues_diffs)), trues_diffs, color="red", linewidth=2
            )

            plt.title(f"averages absolute differences on indexes - {key_ref}")
            plt.xlabel("position in window")
            plt.legend(["ratio", "pred", "true"])
            plt.xticks(np.arange(0, len(norm_pred), 1))
            plt.grid()
            plt.show()



from copyreg import pickle
import json
from matplotlib import rcParams, pyplot as plt
import numpy as np
import pickle

from tensorflow.keras import backend as K
from transformers.modeling_tf_outputs import TFMaskedLMOutput


def extract_results(
    model,
    data,
    masked_positions,
    results_sp: dict,
    results_prot: list,
    conf_matrix: dict,
    inverted_vocab: dict,
    matrix_identity=None,
    matrix_similarity=None,
    matrix_set_similarity=None,
    matrix_cosine_similarity=None,
    start_token=5,
):
    predictions = model.predict(data[0])
    if isinstance(predictions, TFMaskedLMOutput):
        predictions = K.softmax(predictions.logits, axis=-1)
    predicted_tokens = K.argmax(predictions, axis=-1)

    for i, prediction in enumerate(predictions):
        dom_lin = data[3][i]

        argmax_token = int(predicted_tokens[i][masked_positions[i]])
        true_token = data[1][i][masked_positions[i]]
        true_token_seq = inverted_vocab[true_token]

        soft_argmax = float(prediction[masked_positions[i]][argmax_token])
        soft_true = float(prediction[masked_positions[i]][true_token])

        ratio = soft_true / soft_argmax
        hit = 1 if ratio > 0.999 else 0

        results_sp[true_token_seq][dom_lin][0] += hit
        results_sp[true_token_seq][dom_lin][1] += soft_true
        results_sp[true_token_seq][dom_lin][2] += soft_argmax
        results_sp[true_token_seq][dom_lin][3] += 1

        results_prot[0].append(round(soft_true, 5))
        results_prot[1].append(round(soft_argmax, 5))

        conf_matrix[true_token_seq][argmax_token - start_token] += 1

        if matrix_identity:
            results_sp[true_token_seq][dom_lin][4] += matrix_identity[true_token_seq][
                argmax_token - start_token
            ]
        if matrix_similarity:
            results_sp[true_token_seq][dom_lin][5] += matrix_similarity[true_token_seq][
                argmax_token - start_token
            ]
        if matrix_set_similarity:
            results_sp[true_token_seq][dom_lin][6] += matrix_set_similarity[
                true_token_seq
            ][argmax_token - start_token]
        if matrix_cosine_similarity:
            results_sp[true_token_seq][dom_lin][7] += matrix_cosine_similarity[
                true_token_seq
            ][argmax_token - start_token]

    return results_sp, results_prot, conf_matrix


def plot_A(x_sets, y_sets, title, legend=[], alpha=0.1):
    rcParams.update({"figure.autolayout": True})
    plt.figure(figsize=(17, 3.5))
    plt.ion()

    for x, y in zip(x_sets, y_sets):
        plt.scatter(x, y, alpha=alpha)

    plt.title(title)
    plt.xlabel("occurence")
    plt.ylabel("accuracy")
    plt.legend(legend)
    plt.grid()
    plt.show()


def plot_B(accuracies, title):
    rcParams.update({"figure.autolayout": True})
    plt.figure(figsize=(17, 3.5))
    plt.ion()

    plt.hist(accuracies, bins=100)

    plt.title(title)
    plt.xlabel("accuracy")
    plt.ylabel("amount of units")
    plt.grid()
    plt.show()


def show_results_sp(
    results_path,
):
    with open(results_path, "r") as rf:
        data = json.load(rf)

    results_structure = {
        "accuracy": 0,
        "identity": 4,
        "similarity": 5,
        "set_similarity": 6,
    }

    metrics = {
        "accuracy": [],
        "identity": [],
        "similarity": [],
        "set_similarity": [],
    }

    combined_sets_values = list()
    combined_sets_occurances = list()

    for mode, index in results_structure.items():
        values_domains = list()
        occurence_domains = list()
        values_linkers = list()
        occurence_linkers = list()
        values_unknown = list()
        occurence_unknown = list()

        values_concatenated = list()
        occurence_concatenated = list()

        for token in data.keys():
            values_sum = 0
            occ_sum = 0
            for type_dom_lin, type_data in data[token].items():
                occ = int(type_data[3])
                index_values = int(type_data[index])
                value = 0 if not occ else index_values / occ
                if type_dom_lin == "D":
                    values_domains.append(value)
                    occurence_domains.append(occ)
                elif type_dom_lin == "L":
                    values_linkers.append(value)
                    occurence_linkers.append(occ)
                elif type_dom_lin == "U":
                    values_unknown.append(value)
                    occurence_unknown.append(occ)

                values_sum += index_values
                occ_sum += occ
            avg_sum = 0 if not occ_sum else values_sum / occ_sum
            values_concatenated.append(avg_sum)
            occurence_concatenated.append(occ_sum)

        res = (
            sum(values_domains) / sum(occurence_domains)
            if sum(occurence_domains)
            else 0
        )
        metrics[mode].append(res)
        print(f"domains {mode}: {res}")
        res = (
            sum(values_linkers) / sum(occurence_linkers)
            if sum(occurence_linkers)
            else 0
        )
        metrics[mode].append(res)
        print(f"linkers {mode}: {res}")
        res = (
            sum(values_unknown) / sum(occurence_unknown)
            if sum(occurence_unknown)
            else 0
        )
        metrics[mode].append(res)
        print(f"unknown {mode}: {res}")

        plot_A(
            [occurence_domains], [values_domains], f"Tokens {mode}-occurence: domains"
        )
        plot_B(values_domains, f"Tokens {mode} histogram: domains")
        plot_A(
            [occurence_linkers], [values_linkers], f"Tokens {mode}-occurence: linkers"
        )
        plot_B(values_linkers, f"Tokens {mode} histogram: linkers")
        plot_A(
            [occurence_unknown], [values_unknown], f"Tokens {mode}-occurence: unknown"
        )
        plot_B(values_unknown, f"Tokens {mode} histogram: unknown")

        plot_A(
            [occurence_domains, occurence_linkers, occurence_unknown],
            [values_domains, values_linkers, values_unknown],
            f"Tokens {mode}-occurence: comparison",
            ["domains", "linkers", "unknown"],
        )
        plot_A(
            [occurence_concatenated],
            [values_concatenated],
            f"Tokens {mode}-occurence: combined",
        )

        combined_sets_occurances.append(occurence_concatenated)
        combined_sets_values.append(values_concatenated)

    plot_A(
        combined_sets_occurances,
        combined_sets_values,
        f"modes comparison",
        list(results_structure.keys()),
    )

    plot_A(
        [range(3) for _ in range(len(metrics))],
        list(metrics.values()),
        f"metrics",
        list(results_structure.keys()),
        alpha=1,
    )


def show_results_proteins(
    results_path,
):
    accuracies = list()
    with open(results_path, "rb") as rf:
        for line in rf.readlines():
            line = line.decode().split(";")

            gt = [(float(val)) for val in line[1].split(",")]
            argmax = [(float(val)) for val in line[2].split(",")]

            hits = [int(t / a) for t, a in zip(gt, argmax)]
            accuracies.append(sum(hits) / len(hits))

        plot_B(accuracies, "Proteins accuracy histogram")

    # for values in data.values():
    #     gt = values[0]
    #     argmax = values[1]

    #     hits = [int(t / a) for t, a in zip(gt, argmax)]

    #     accuracies.append(sum(hits) / len(hits))

    # plot_B(accuracies, "Proteins accuracy histogram")



import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras.metrics import Metric
from tensorflow.keras.losses import SparseCategoricalCrossentropy

from ..constants import TOKENIZER

CE = SparseCategoricalCrossentropy()


class CrossEntropyMasked(Metric):
    def __init__(self, name="accuracy_masked"):
        super(CrossEntropyMasked, self).__init__()
        self.accuracy = self.add_weight(name=name, initializer="zeros")
        self.tics = self.add_weight(name=name, initializer="zeros")

    def update_state(self, y_true, y_pred, sample_weight=None):
        x = K.cast(y_pred[:, :, 0], 'int64')
        y_pred = y_pred[:, :, 1:]
        y_true = K.cast(y_true, "int64")

        is_masked = K.equal(x, TOKENIZER.mask_token_id)

        y_true = tf.boolean_mask(y_true, is_masked)
        y_pred = tf.boolean_mask(y_pred, is_masked)

        loss = CE(y_true, y_pred)

        self.accuracy.assign_add(loss)
        self.tics.assign_add(1.)

    def result(self):
        return self.accuracy / self.tics

    def reset(self):
        self.accuracy.assign(0.)
        self.tics.assign(0.)



import json

import numpy as np


# GROUPS_REVERSE = json.load(open('prot2vec/configs/helper_groups_reverse.json', 'r', encoding='utf-8'))
# GROUPS_REVERSE = json.load(open('../configs/helper_groups_reverse.json', 'r', encoding='utf-8'))


def mask_by_helper_tokens(tokens_in, bool_mask, help_distrib, tokenizer):
    to_replace = tokens_in[bool_mask]

    help_distrib = np.array([help_distrib['10'], help_distrib['100'], help_distrib['1000'], help_distrib['none']])
    selected_helps = np.array(['10', '100', '1000', 'none'])[np.random.choice(4, len(to_replace), p=help_distrib)]

    replacement = []
    for j in range(len(to_replace)):
        selected_help = selected_helps[j]
        orig_token = to_replace[j]
        mask_token = tokenizer.mask_token_id if selected_help == 'none' else GROUPS_REVERSE[str(orig_token)][
            selected_help]
        replacement.append(mask_token)

    tokens_in[bool_mask] = np.array(replacement)
    return tokens_in


def mask_data(data, mask_prob, help_distrib, tokenizer):
    bool_mask = np.random.rand(data.shape[0] * data.shape[1]) < mask_prob
    bool_mask = bool_mask.reshape(data.shape)
    bool_mask[:, 0] = False
    bool_mask[:, -1] = False

    data = mask_by_helper_tokens(data, bool_mask, help_distrib, tokenizer)
    return data, bool_mask


def vectorize_batch(batch, mask_prob, help_distrib, tokenizer):
    tokens_in = np.concatenate(batch, axis=0)
    tokens_out = tokens_in.copy()
    if mask_prob > 1e-6:
        tokens_in, mask = mask_data(tokens_in, mask_prob, help_distrib, tokenizer)

    return tokens_in, tokens_out, mask


def get_help_scheduler(helper_mode):
    if helper_mode == '10':
        help_sched = lambda x: {'10': 1., '100': 0., '1000': 0., 'none': 0.}
    elif helper_mode == '100':
        help_sched = lambda x: {'10': 0., '100': 1., '1000': 0., 'none': 0.}
    elif helper_mode == '1000':
        help_sched = lambda x: {'10': 0., '100': 0., '1000': 1., 'none': 0.}
    elif helper_mode == 'none':
        help_sched = lambda x: {'10': 0., '100': 0., '1000': 0., 'none': 1.}
    elif helper_mode == 'continuous':
        help_sched = lambda x: {'10': (x < 1 / 3) * (1 - 3 * x),
                                '100': (x < 1 / 3) * (3 * x) + (1 / 3 <= x < 2 / 3) * (2 - 3 * x),
                                '1000': (1 / 3 <= x < 2 / 3) * (3 * x - 1) + (2 / 3 <= x) * (3 - 3 * x),
                                'none': (2 / 3 <= x) * (3 * x - 2)}
    else:
        raise Exception(f'helper_mode \"{helper_mode}\" is not supported.')

    return help_sched


def data_gen(data_path, tokenizer, batch_size, max_len, mask_prob=0.15, overfit_batch=False, helper_mode='none',
             once=False, session_steps=-1, skip_shorter_than=None):
    if skip_shorter_than is None:
        skip_shorter_than = max_len
    # Setup helper scheduler
    help_sched = get_help_scheduler(helper_mode)
    step = 0

    if helper_mode == 'continuous':
        assert session_steps > 0

    while True:
        batch = []
        with open(data_path, "r", encoding="utf-8") as fr:
            fr.readline()
            for line in fr:
                items = line.strip().split("\t")
                id, seq_id, seq = items

                tokens = tokenizer(seq)['input_ids']
                if len(tokens) < skip_shorter_than:
                    continue

                offset = np.random.randint(0, len(tokens) - max_len + 1)
                tokens = tokens[offset:offset + max_len]
                tokens[0] = tokenizer.cls_token_id
                tokens[-1] = tokenizer.eos_token_id

                batch.append(np.array(tokens)[None, :])

                if len(batch) >= batch_size:
                    v_batch = vectorize_batch(batch, mask_prob, help_sched(min(step / session_steps, 1.)), tokenizer)
                    if overfit_batch:
                        while True:
                            yield v_batch
                    yield v_batch
                    step += 1
                    del batch[:]

        if once:
            v_batch = vectorize_batch(batch, mask_prob, help_sched(min(step / session_steps, 1.)), tokenizer)
            yield v_batch
            break


def random_iter(tokenizer, batch_size, max_len, mask_prob, helper_mode='none', session_steps=-1):
    with open('prot2vec/configs/1st_order_random_dict.json', 'r', encoding='utf-8') as fr:
        sp_probs = json.load(fr)
    sp_probs = sorted([(int(k), v) for k, v in sp_probs.items()], key=lambda x: x[1])
    sp_ids = np.array([k for k, v in sp_probs])
    probs = [v for k, v in sp_probs]

    help_sched = get_help_scheduler(helper_mode)
    if helper_mode == 'continuous':
        assert session_steps > 0
    step = 1

    while True:
        selected_indexes = np.random.choice(len(sp_ids), max_len * batch_size, p=probs)
        tokens_in = sp_ids[selected_indexes]
        tokens_in = tokens_in.reshape((batch_size, max_len))
        tokens_in[:, 0] = 0
        tokens_in[:, -1] = 2

        tokens_out = tokens_in.copy()

        if mask_prob > 1e-6:
            tokens_in = mask_data(tokens_in, mask_prob, help_sched(min(step / session_steps, 1.)), tokenizer)

        yield tokens_in, tokens_out
        step += 1



### TO SETUP ###
FINAL_MODELS_DIR = '/auto/plzen4-ntis/projects/BioML/tomas_lukas/final_models/'
RESULTS_DIR = '/auto/plzen4-ntis/projects/BioML/tomas_lukas/results/'
DATA_DIR = '/storage/plzen4-ntis/projects/BioML/datasets/proteoconv/'
VAL_DIR = '/auto/plzen4-ntis/projects/BioML/tomas_lukas/data/'
STATS_DIR = '/auto/plzen4-ntis/projects/BioML/tomas_lukas/statistics'

### CONCATENATED ###
REFPROT_TSV = VAL_DIR + 'refprot_random_taxfree.tsv'
TAX_TSV = VAL_DIR + 'tax_ds_1000_prots_from_each_order.tsv'

### OTHER ###
TOT_TRAIN_PROT = 47082110



from tensorflow.keras.layers import Add, Dropout, BatchNormalization, Activation, Conv1D, Conv1DTranspose


def conv1D_identity_block(x, kernels, kernel_size, dropout=0.1, activation='relu'):
    x_skip = x
    # Layer 1
    x = Conv1D(kernels, kernel_size, padding='same')(x)
    x = BatchNormalization()(x)
    x = Activation(activation)(x)
    if dropout > 0.0001:
        x = Dropout(dropout)(x)
    # Layer 2
    x = Conv1D(kernels, kernel_size, padding='same')(x)
    x = BatchNormalization()(x)
    # Add Residue
    x = Add()([x, x_skip])
    x = Activation(activation)(x)
    if dropout > 0.0001:
        x = Dropout(dropout)(x)
    return x


def conv1D_conv_block(x, kernels, kernel_size, dropout=0.1, activation='relu'):
    x_skip = x
    # Layer 1
    x = Conv1D(kernels, kernel_size, padding='same', strides=2)(x)
    x = BatchNormalization()(x)
    x = Activation(activation)(x)
    if dropout > 0.0001:
        x = Dropout(dropout)(x)
    # Layer 2
    x = Conv1D(kernels, kernel_size, padding='same')(x)
    x = BatchNormalization()(x)

    x_skip = Conv1D(kernels, 1, strides=2)(x_skip)
    # Add Residue
    x = Add()([x, x_skip])
    x = Activation(activation)(x)
    if dropout > 0.0001:
        x = Dropout(dropout)(x)
    return x


def conv1D_deconv_block(x, kernels, kernel_size, dropout=0.1, activation='relu'):
    x_skip = x
    # Layer 1
    x = Conv1DTranspose(kernels, kernel_size, padding='same', strides=2)(x)
    x = BatchNormalization()(x)
    x = Activation(activation)(x)
    if dropout > 0.0001:
        x = Dropout(dropout)(x)
    # Layer 2
    x = Conv1D(kernels, kernel_size, padding='same')(x)
    x = BatchNormalization()(x)

    x_skip = Conv1DTranspose(kernels, 1, strides=2)(x_skip)
    # Add Residue
    x = Add()([x, x_skip])
    x = Activation(activation)(x)
    if dropout > 0.0001:
        x = Dropout(dropout)(x)
    return x



import datetime
import os
import zlib
from pathlib import Path
import os

import pandas as pd
from tensorflow.keras.callbacks import CSVLogger, TensorBoard, Callback, ModelCheckpoint

from ..constants import RESULTS_DIR


def get_logging_callbacks(params, experiment_id, param_id, experiment_name=None, results_path=RESULTS_DIR):
    logger = ExperimentLogger(params, experiment_id, param_id, experiment_name, results_path)
    csv_logger = CSVLogger(logger.path / 'history.tsv', separator='\t', append=True)
    tensorboard = TensorBoard(log_dir=logger.tensorboard_path)

    checkpoint_n = len(os.listdir(logger.checkpoints_path)) + 1
    checkpoint_n = str(checkpoint_n).zfill(5)
    checkpoint_n = f'{checkpoint_n}_checkpoint.h5'

    checkpointer = ModelCheckpoint(
        logger.checkpoints_path / checkpoint_n, monitor='loss', verbose=0, save_best_only=False,
        save_weights_only=True, mode='auto', save_freq='epoch'
    )

    return [logger, csv_logger, tensorboard, checkpointer]


class ExperimentLogger(Callback):
    def __init__(self, params, experiment_id, param_id, experiment_name=None, results_path=RESULTS_DIR):
        super(ExperimentLogger, self).__init__()
        if experiment_name is None:
            raise Exception('Please specify experiment name.')
        results_path = Path(results_path)

        # Attributes init
        self.params = params
        self.experiment_name = experiment_name
        self.params_hash = self.__get_hash_params(params)
        self.history = dict()
        self.time_stamp = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

        # Create directory for the logs
        self.root_path = results_path
        self.path = results_path / '_flat' / experiment_id
        if not os.path.exists(self.path):
            os.makedirs(self.path)
        self.path = self.path / param_id
        if not os.path.exists(self.path):
            os.makedirs(self.path)
        # self.path = self.path / str(self.params_hash)
        # if not os.path.exists(self.path):
        #     os.makedirs(self.path)
        # self.path = self.path / self.time_stamp
        # if not os.path.exists(self.path):
        #     os.makedirs(self.path)
        if not os.path.exists(self.path / 'models'):
            os.makedirs(self.path / 'models')
        self.model_path = self.path / 'models'
        if not os.path.exists(self.path / 'models' / 'checkpoints'):
            os.makedirs(self.path / 'models' / 'checkpoints')
        self.checkpoints_path = self.path / 'models' / 'checkpoints'
        if not os.path.exists(self.path / 'tensorboard'):
            os.makedirs(self.path / 'tensorboard')
        self.tensorboard_path = self.path / 'tensorboard'
        if not os.path.exists(self.root_path / '_archive'):
            os.makedirs(self.root_path / '_archive')
        self.archive_path = self.root_path / '_archive'

        self.update_result_table()

    def __get_params_string(self, params):
        keys = sorted([str(key) for key in params])
        values = [str(params[key]) for key in keys]
        return ';'.join([':'.join([keys[i], values[i]]) for i in range(len(keys))])

    def __get_hash_params(self, params):
        string_params = self.__get_params_string(params)
        return zlib.adler32(bytes(string_params, 'utf-8'))

    def update_result_table(self):
        results_table = pd.read_csv(self.root_path / 'results.tsv', delimiter='\t', index_col=False)
        results_table.to_csv(self.archive_path / str(self.time_stamp + '_results.tsv.gz'), sep='\t',
                             header=True,
                             index=False, compression='gzip')

        new_row = {
            'Experiment_Name': self.experiment_name,
            'Time_Stamp': self.time_stamp,
            'Params_Hash': self.__get_hash_params(self.params),
            'Params_String': self.__get_params_string(self.params)
        }
        results_table = results_table.append(new_row, ignore_index=True)
        results_table.to_csv(self.root_path / 'results.tsv', sep='\t', header=True, index=False)



from tensorflow.keras import backend as K
from tensorflow.keras.metrics import Metric, sparse_categorical_accuracy


class AccuracyWithoutPad(Metric):
    def __init__(self, name='accuracy_without_pad'):
        super(AccuracyWithoutPad, self).__init__()
        self.accuracy = self.add_weight(name=name, initializer='zeros')
        self.tics = self.add_weight(name='tics', initializer='zeros')

    def update_state(self, y_true, y_pred, x=None, sample_weight=None):
        y_true = y_true[:, 1:-1]
        y_pred = y_pred[:, 1:-1]
        acc = sparse_categorical_accuracy(y_true, y_pred)
        acc = K.mean(acc)

        self.accuracy.assign_add(acc)
        self.tics.assign_add(1.)

    def result(self):
        return self.accuracy / self.tics

    def reset(self):
        self.accuracy.assign(0.)
        self.tics.assign(0.)



import json

import numpy as np

AA_DICT = {'?': 0, '<': 1, '>': 2,
           'A': 3, 'C': 4, 'D': 5, 'E': 6, 'F': 7,
           'G': 8, 'H': 9, 'I': 10, 'K': 11, 'L': 12,
           'M': 13, 'N': 14, 'P': 15, 'Q': 16, 'R': 17,
           'S': 18, 'T': 19, 'V': 20, 'W': 21, 'Y': 22}

AA_DICT_REV = {AA_DICT[key]: key for key in AA_DICT}

AA_DISTRIBUTION = np.load('prot2vec/configs/probab_distribution.npy')


def tokenize(seq):
    digit_seq = []
    for aa in seq:
        if aa not in AA_DICT:
            print(f'Unknown amino acid letter \"{aa}\"')
            return None
        digit_seq.append(AA_DICT[aa])

    digit_seq = np.array(digit_seq)

    digit_seq[0] = 1
    digit_seq[-1] = 2

    return digit_seq[None, :]


def mask_data(data, mask_prob):
    bool_mask = np.random.rand(data.shape[0] * data.shape[1]) < mask_prob
    bool_mask = bool_mask.reshape(data.shape)
    bool_mask[:, 0] = False
    bool_mask[:, -1] = False

    to_replace = data[bool_mask]

    replacement = np.zeros(to_replace.shape, dtype='int32')

    data[bool_mask] = replacement

    return data, bool_mask


def vect_batch(batch, mask_prob):
    x = np.concatenate(batch, axis=0)
    y = x.copy()

    x, mask = mask_data(x, mask_prob)

    return x, y, mask


def random_iter(seq_len, masked_size, batch_size, overfit_batch=False):
    while True:
        seq_in_ = np.random.choice(len(AA_DISTRIBUTION), batch_size * seq_len, p=AA_DISTRIBUTION)

        seq_in = np.zeros((batch_size * seq_len, len(AA_DISTRIBUTION)))
        arang = np.arange(0, seq_in.shape[0])

        seq_in[arang, seq_in_] = 1.0
        seq_in = seq_in.reshape((batch_size, seq_len, len(AA_DISTRIBUTION)))

        m_offset = np.random.randint(0, seq_len - masked_size + 1)
        seq_out = seq_in[:, m_offset:m_offset + masked_size, :].copy()
        seq_in[:, m_offset:m_offset + masked_size, :] = 0.0

        if overfit_batch:
            while True:
                yield seq_in, seq_out

        yield seq_in, seq_out


def ds_iter(src_file, seq_len, mask_prob, batch_size):
    batch = []
    while True:
        with open(src_file, "r", encoding="utf-8") as fr:
            for line in fr:
                _, _, seq = line[:-1].split('\t')
                prot_len = len(seq[:-1])
                if prot_len <= seq_len:
                    continue

                offset = np.random.randint(0, prot_len - seq_len + 1)
                seq = seq[offset:offset + seq_len]

                seq_in = tokenize(seq)

                if seq_in is None:
                    continue

                batch.append(seq_in)

                if len(batch) >= batch_size:
                    v_batch = vect_batch(batch, mask_prob)
                    yield v_batch
                    del batch[:]



import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras.metrics import Metric

from ..constants import TOKENIZER


class AccuracyMasked(Metric):
    def __init__(self, name="accuracy_masked"):
        super(AccuracyMasked, self).__init__()
        self.accuracy = self.add_weight(name=name, initializer="zeros")
        self.tics = self.add_weight(name='tics', initializer="zeros")

    def update_state(self, y_true, y_pred, sample_weight=None):
        x = K.cast(y_pred[:, :, 0], 'int64')
        y_pred = y_pred[:, :, 1:]

        y_true = K.cast(y_true, "int64")
        y_pred = K.argmax(y_pred, axis=-1)
        eq_out = K.equal(y_true[:, 1:-1], y_pred[:, 1:-1])
        eq_masked = K.equal(x[:, 1:-1], TOKENIZER.mask_token_id)

        logic_and = tf.logical_and(eq_out, eq_masked)
        acc_of_masked = K.sum(K.cast(logic_and, "int32")) / K.sum(K.cast(eq_masked, "int32"))
        acc_of_masked = K.cast(acc_of_masked, "float32")
        self.accuracy.assign_add(acc_of_masked)
        self.tics.assign_add(1.)

    def result(self):
        return self.accuracy / self.tics

    def reset(self):
        self.accuracy.assign(0.)
        self.tics.assign(0.)



import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras.metrics import Metric


def update_metric(x_prev, x_now, i):
    return (1 / (i + 1)) * (i * x_prev + x_now)


class SampleAccHist(Metric):
    def __init__(self, name='sample_accuracy_histogram'):
        super(SampleAccHist, self).__init__()
        self.metric_name = name
        self.metric = []
        self.tics = self.add_weight('tics', initializer="zeros")

    def update_state(self, y_true, y_pred, x=None, sample_weight=None):
        y_pred = tf.argmax(y_pred, axis=-1, output_type=tf.int32)
        equals = tf.cast(y_pred == y_true, tf.int32)

        linspac = tf.range(y_pred.shape[1] + 1)

        eq_sum = K.sum(equals, axis=1)

        eq_sum = tf.repeat(eq_sum[:, None], linspac.shape[0], axis=1)
        linspac = tf.repeat(linspac[None, :], eq_sum.shape[0], axis=0)

        hits_count_histogram = K.cast(eq_sum == linspac, 'float32')
        hits_count_histogram = K.sum(hits_count_histogram, axis=0)
        hits_count_histogram = hits_count_histogram / K.sum(hits_count_histogram)

        for i in range(hits_count_histogram.shape[0]):
            if i >= len(self.metric):
                self.metric.append(self.add_weight(name=self.metric_name + str(i), initializer="zeros"))
            self.metric[i].assign(update_metric(self.metric[i], hits_count_histogram[i], self.tics))

        self.tics.assign_add(1.)

    def result(self):
        return self.metric

    def reset(self):
        for i in range(len(self.metric)):
            self.metric[i].assign(0.)
        self.tics.assign(0.)



from tensorflow import keras
import tensorflow.keras.backend as K
import tensorflow as tf


class IOEmbedding(keras.layers.Embedding):
    def __init__(self,
                 initializer='zeros',
                 regularizer=None,
                 constraint=None,
                 **kwargs):
        super(IOEmbedding, self).__init__(**kwargs)
        self.supports_masking = True
        self.initializer = keras.initializers.get(initializer)
        self.regularizer = keras.regularizers.get(regularizer)
        self.constraint = keras.constraints.get(constraint)
        self.bias = None

    def get_config(self):
        config = {
            'initializer': keras.initializers.serialize(self.initializer),
            'regularizer': keras.regularizers.serialize(self.regularizer),
            'constraint': keras.constraints.serialize(self.constraint),
        }
        base_config = super(IOEmbedding, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def build(self, input_shape):
        self.bias = self.add_weight(
            shape=(self.input_dim,),
            initializer=self.initializer,
            regularizer=self.regularizer,
            constraint=self.constraint,
            name='bias',
        )
        super(IOEmbedding, self).build(input_shape)

    def compute_mask(self, inputs, inverse=False, mask=None):
        if not isinstance(inputs, list) or len(inputs) != 2:
            return super(IOEmbedding, self).compute_mask(inputs, mask)
        else:
            token_mask = K.not_equal(inputs[1], 0)
            return token_mask

    def call(self, inputs, mask=None, inverse=False, **kwargs):
        if not inverse:
            return super(IOEmbedding, self).call(inputs)
        else:
            if not isinstance(inputs, list):
                inputs = [inputs]
            
            return tf.tensordot(inputs[0], self.embeddings, [[2], [1]])+self.bias



import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras.metrics import Metric


class MeanTopK(Metric):
    def __init__(self, is_weighted=False, name='mean_top_k'):
        super(MeanTopK, self).__init__()
        self.mean_top_k_sum = self.add_weight(name=name, initializer='zeros', dtype='float64')
        self.ticks = self.add_weight(name='tics', initializer='zeros', dtype='float64')
        self.is_weighted = is_weighted

    def update_state(self, y_true, y_pred, sample_weight=None, crop_edges=False):
        if crop_edges:
            y_true = y_true[:, 1:-1]
            y_pred = y_pred[:, 1:-1]

        y_top = tf.gather(y_pred, y_true, batch_dims=2)
        y_top_rep = tf.repeat(y_top[:, :, None], y_pred.shape[2], axis=-1)

        higher_than_tru = K.cast(y_pred > y_top_rep, dtype='float64')

        top_k = tf.reduce_sum(higher_than_tru, axis=-1)

        if self.is_weighted:
            top_k = top_k * K.cast(sample_weight, 'float64')
            top_k_mean = K.sum(top_k) / K.cast(K.sum(sample_weight), 'float64')
        else:
            top_k_mean = K.mean(top_k)

        self.mean_top_k_sum.assign_add(top_k_mean)
        self.ticks.assign_add(1.)

    def result(self):
        return self.mean_top_k_sum / self.ticks

    def reset(self):
        self.accuracy.assign(0.)
        self.tics.assign(0.)



import math

from keras_pos_embd import PositionEmbedding
from tensorflow.keras.layers import Input, Dense, Embedding
from tensorflow.keras.models import Model

from prot2vec.layers.transformer_block import TransformerBlock


def bert_model(seq_len, embedding_size, vocab_size, n_layers, n_heads, ff_dim, with_hints=True):
    if with_hints:
        vocab_size = math.ceil(vocab_size * 1.11111111)

    input_layer = Input(shape=(seq_len,), dtype='int32', name='input_tokens')

    x = Embedding(vocab_size, embedding_size)(input_layer)
    x = PositionEmbedding(input_dim=seq_len, output_dim=embedding_size, mode=PositionEmbedding.MODE_ADD)(x)
    # x = TrigPosEmbedding(input_dim=seq_len, output_dim=embedding_size, mode=TrigPosEmbedding.MODE_ADD)(x)

    for i in range(n_layers):
        x = TransformerBlock(embed_dim=embedding_size, num_heads=n_heads, ff_dim=ff_dim)(x)

    output_layer = Dense(vocab_size, activation='softmax')(x)
    model = Model(inputs=[input_layer], outputs=[output_layer], name="prot_bert")

    return model


if __name__ == '__main__':
    b = bert_model(24, 256, 8000, 4, 4, 1024)
    print()



from setuptools import setup, find_packages

VERSION = '0.0.24'
DESCRIPTION = 'Vectorization and Annotation of Protein Sequences'
LONG_DESCRIPTION = 'A package that allows to vectorize and annotate protein sequences.'

# Setting up
setup(
    name="prot2vec",
    version=VERSION,
    author="Tomas Honzik",
    author_email="<tomas.honziik@seznam.cz>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=[],
    keywords=['python', 'protein'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



import json
import os

from keras_pos_embd import PositionEmbedding, TrigPosEmbedding
from tensorflow.keras.layers import Input, Lambda, GaussianNoise, RepeatVector, LayerNormalization, Activation, Dense, \
    Embedding
from tensorflow.keras.models import Model, load_model
from transformers.optimization_tf import WarmUp

from ..layers.ioembedding import IOEmbedding
from ..layers.transformer_block import TransformerBlock


def get_model(max_len, dim, num_layers, num_heads, ff_dim, vocab_siz, model_dir=None, train_from_scratch=True,
              short_circuit_reduction=False, gaussian_std=None, with_hints=True):
    input_in = Input(shape=(max_len, dim), dtype='float32')

    # interim = TrigPosEmbedding(mode=TrigPosEmbedding.MODE_ADD)(input_in)
    interim = PositionEmbedding(input_dim=max_len, output_dim=dim, mode=PositionEmbedding.MODE_ADD)(input_in)

    for i in range(num_layers):
        interim = TransformerBlock(embed_dim=dim, num_heads=num_heads, ff_dim=ff_dim)(interim)
    encoded_full = interim

    encoded_input = Input(shape=(dim,), dtype='float32')
    encoded_unrolled = RepeatVector(max_len)(encoded_input)

    interim = PositionEmbedding(input_dim=max_len, output_dim=dim, mode=PositionEmbedding.MODE_ADD)(encoded_unrolled)
    interim = LayerNormalization(epsilon=1e-6)(interim)

    for i in range(num_layers):
        interim = TransformerBlock(embed_dim=dim, num_heads=num_heads, ff_dim=ff_dim)(interim)
    decoded_output = interim

    if model_dir is not None and os.path.isdir(model_dir) and not train_from_scratch:
        print("Loading model from", model_dir)
        model = load_model(os.path.join(model_dir, "model"),
                           custom_objects={"WarmUp": WarmUp, "TransformerBlock": TransformerBlock,
                                           "IOEmbedding": IOEmbedding})

        with open(f"{model_dir}/config.json", "r", encoding="utf-8") as fr:
            TRAIN_STATE = json.load(fr)
            INITIAL_EPOCH = TRAIN_STATE["INITIAL_EPOCH"]
        print(f"Training will start from state: {TRAIN_STATE}")
    else:
        print("Initializing a new model from scratch")

        if with_hints:
            vocab_siz = int(vocab_siz * 1.2)

        emb_layer = IOEmbedding(input_dim=vocab_siz, output_dim=dim, name="ioembedding")
        emb_layer_2 = Embedding(input_dim=vocab_siz, output_dim=dim)

        embedder = Model(input_in, encoded_full, name="embedder")

        # reduction = Model(ae_input, encoded_reduced, name="reduction")
        # expansion = Model(ae_reduced, encoded_expanded, name="expansion")
        deembedder = Model(encoded_input, decoded_output, name="deembedder")
        TRAIN_STATE = {}

        model_in = Input(shape=(max_len,), dtype='int32', name="input_tokens")
        model_mask = Input(shape=(max_len,), dtype='int32', name="input_mask")

        x = emb_layer_2(model_in)
        x = embedder(x)

        # masked_out = emb_layer([x, model_mask], inverse=True)
        masked_out = Dense(vocab_siz, 'softmax')(x)

        x = Lambda((lambda i: i[:, 0]), name="bos_vector")(x)

        if not short_circuit_reduction:
            x = reduction(x)
        if gaussian_std is not None:
            x = GaussianNoise(gaussian_std)(x)
        if not short_circuit_reduction:
            x = expansion(x)
        x = deembedder(x)

        out = emb_layer(x, inverse=True)
        out = Activation('softmax')(out)
        model = Model(inputs={"tokens": model_in, "mask": model_mask}, outputs={"deemb": out, "emb": masked_out})

    return model


def get_model_2(max_len, dim, num_layers, num_heads, ff_dim, vocab_siz):
    model_mask = Input(shape=(max_len,), dtype='int32', name="input_mask")
    input_in = Input(shape=(max_len, dim), dtype='float32')

    # interim = TrigPosEmbedding(mode=TrigPosEmbedding.MODE_ADD)(input_in)
    interim = PositionEmbedding(input_dim=max_len, output_dim=dim, mode=PositionEmbedding.MODE_ADD)(input_in)

    for i in range(num_layers):
        interim = TransformerBlock(embed_dim=dim, num_heads=num_heads, ff_dim=ff_dim)(interim)
    encoded_full = interim

    vocab_siz = int(vocab_siz * 1.2)

    # emb_layer = IOEmbedding(input_dim=vocab_siz, output_dim=dim, name="ioembedding")
    emb_layer_2 = Embedding(input_dim=vocab_siz, output_dim=dim)

    embedder = Model(input_in, encoded_full, name="embedder")

    model_in = Input(shape=(max_len,), dtype='int32', name="input_tokens")

    x = emb_layer_2(model_in)
    x = embedder(x)

    masked_out = Dense(vocab_siz, 'softmax')(x)

    model = Model(inputs={"tokens": model_in, "mask": model_mask}, outputs={"deemb": masked_out, "emb": masked_out})

    return model


def get_model_3(max_len, dim, num_layers, num_heads, ff_dim, vocab_siz):
    vocab_siz = int(vocab_siz * 1.2)

    ### Embedder
    input_in = Input(shape=(max_len, dim), dtype='float32')
    interim = PositionEmbedding(input_dim=max_len, output_dim=dim, mode=PositionEmbedding.MODE_ADD)(input_in)
    for i in range(num_layers):
        interim = TransformerBlock(embed_dim=dim, num_heads=num_heads, ff_dim=ff_dim)(interim)
    encoded_full = interim
    embedder = Model(input_in, encoded_full, name="embedder")
    ###

    model_in = Input(shape=(max_len,), dtype='int32', name="input_tokens")
    x = Embedding(input_dim=vocab_siz, output_dim=dim)(model_in)
    x = embedder(x)
    masked_out = Dense(vocab_siz, 'softmax')(x)

    model = Model(inputs=[model_in], outputs=[masked_out])

    return model


if __name__ == '__main__':
    test_obj = get_model(24, 256, 4, 4, 1024, 8000)
    print()



import numpy as np


def mask_by_helper_tokens(tokens_in, bool_mask, hint_callback, tokenizer):
    to_replace = tokens_in[bool_mask]

    p = hint_callback.get_distribution()
    selected_helps = np.array(['10', '100', '1000', 'none'])[np.random.choice(4, len(to_replace), p=p)]

    replacement = []
    for j in range(len(to_replace)):
        selected_help = selected_helps[j]
        orig_token = to_replace[j]
        mask_token = tokenizer.mask_token_id if selected_help == 'none' else hint_callback.hints[str(orig_token)][
            selected_help]
        replacement.append(mask_token)

    tokens_in[bool_mask] = np.array(replacement)
    return tokens_in


def mask_data(data, mask_prob, hint_callback, tokenizer):
    bool_mask = np.random.rand(data.shape[0] * data.shape[1]) < mask_prob
    bool_mask = bool_mask.reshape(data.shape)
    bool_mask[:, 0] = False
    bool_mask[:, -1] = False

    data = mask_by_helper_tokens(data, bool_mask, hint_callback, tokenizer)
    return data, bool_mask


def vectorize_batch(batch, mask_prob, hint_callback, tokenizer):
    tokens_in = np.concatenate(batch, axis=0)
    tokens_out = tokens_in.copy()
    mask = np.ones(tokens_in.shape)
    if mask_prob > 1e-6:
        tokens_in, mask = mask_data(tokens_in, mask_prob, hint_callback, tokenizer)

    # Make soft zeros mask
    mask = mask.astype('float32') + 1e-5

    return tokens_in, tokens_out, mask


def data_gen_termostat(data_path, tokenizer, batch_size, max_len, hint_callback, mask_prob=0.15):
    while True:
        with open(data_path, "r", encoding="utf-8") as fr:
            batch = []
            for line in fr:
                items = line.strip().split("\t")
                id, seq_id, seq = items

                tokens = tokenizer(seq)['input_ids']
                if len(tokens) < max_len:
                    continue

                offset = np.random.randint(0, len(tokens) - max_len + 1)
                tokens = tokens[offset:offset + max_len]
                tokens[0] = tokenizer.cls_token_id
                tokens[-1] = tokenizer.eos_token_id

                batch.append(np.array(tokens)[None, :])

                if len(batch) >= batch_size:
                    v_batch = vectorize_batch(batch, mask_prob, hint_callback, tokenizer)
                    yield v_batch
                    del batch[:]



import numpy as np
import json


class SentencePieceChangingHints(object):
    def __init__(self, probs_path, vocab_size, hint_size=10):
        with open(probs_path, 'r', encoding='utf-8') as fr:
            sp_probs = json.load(fr)

        sp_probs = sorted([(int(key), sp_probs[key]) for key in sp_probs], key=lambda x: x[0])

        sp_ids = [x[0] for x in sp_probs]
        sp_probs = [x[1] for x in sp_probs]
        sp_probs_cop = sp_probs.copy()
        sp_ids_cop = sp_ids.copy()

        groups = [[]]

        while len(sp_ids) > 0:
            if len(groups[-1]) >= hint_size:
                groups.append([])

            selected = np.random.choice(len(sp_ids), 1, p=sp_probs)
            assert len(selected == 1)
            selected = selected[0]

            groups[-1].append(sp_ids[selected])

            sp_ids.pop(selected)
            sp_probs.pop(selected)
            sp_probs = np.array(sp_probs)
            sp_probs = list(sp_probs / np.sum(sp_probs))

        self.hints = dict()
        for group_i, group in enumerate(groups):
            for tok in group:
                self.hints[tok] = vocab_size + group_i

    def __mask_by_helper_tokens(self, tokens_in, bool_mask, tokenizer):
        to_replace = tokens_in[bool_mask]

        replacement = [self.hints[tok] for tok in to_replace]

        tokens_in[bool_mask] = np.array(replacement)
        return tokens_in

    def __mask_data(self, data, mask_prob, tokenizer):
        bool_mask = np.random.rand(data.shape[0] * data.shape[1]) < mask_prob
        bool_mask = bool_mask.reshape(data.shape)
        bool_mask[:, 0] = False
        bool_mask[:, -1] = False

        data = self.__mask_by_helper_tokens(data, bool_mask, tokenizer)
        return data, bool_mask

    def __vectorize_batch(self, batch, mask_prob, tokenizer):
        tokens_in = np.concatenate(batch, axis=0)
        tokens_out = tokens_in.copy()
        mask = np.ones(tokens_in.shape)
        if mask_prob > 1e-6:
            tokens_in, mask = self.__mask_data(tokens_in, mask_prob, tokenizer)

        # Make soft zeros mask
        mask = mask.astype('float32') + 1e-5

        return tokens_in, tokens_out, mask

    def data_gen(self, data_path, tokenizer, batch_size, max_len, mask_prob=0.15):
        while True:
            with open(data_path, "r", encoding="utf-8") as fr:
                batch = []
                for line in fr:
                    items = line.strip().split("\t")
                    id, seq_id, seq = items

                    tokens = tokenizer(seq)['input_ids']
                    if len(tokens) < max_len:
                        continue

                    offset = np.random.randint(0, len(tokens) - max_len + 1)
                    tokens = tokens[offset:offset + max_len]
                    tokens[0] = tokenizer.cls_token_id
                    tokens[-1] = tokenizer.eos_token_id

                    batch.append(np.array(tokens)[None, :])

                    if len(batch) >= batch_size:
                        v_batch = self.__vectorize_batch(batch, mask_prob, tokenizer)
                        yield v_batch
                        del batch[:]
